mod readmap_helper;
use crossterm::event::KeyCode;
use readmap::types::MemoryMap;
use std::path::PathBuf;
use tool_helper::Error;
use ratatui::{
    layout::{Alignment, Constraint, Direction, Layout, Rect},
    style::{Color, Modifier, Style},
    text::{Line, Span},
    widgets::{Block, Borders, BorderType, Gauge, List, ListItem, ListState, Paragraph, Tabs}
};

pub type EventReceiver = std::sync::mpsc::Receiver<Event<crossterm::event::KeyEvent>>;
pub type Terminal      = ratatui::Terminal<ratatui::backend::CrosstermBackend<std::io::Stdout>>;
type ConsoleFrame<'a>  = ratatui::Frame<'a, ratatui::backend::CrosstermBackend<std::io::Stdout>>;

pub enum Event<I> {
    Input(I),
    Tick,
    ForceRender,
}

pub enum UIState {
    Alive,
    Render,
    Terminated
}

pub enum MenuSelection {
    Stats,
    Quit,
}

impl MenuSelection {
    fn next(&self) -> MenuSelection {
        match self {
            Self::Stats => Self::Quit,
            Self::Quit  => Self::Stats
        }
    }

    fn prev(&self) -> MenuSelection {
        self.next()
    }

    fn get_id(&self) -> usize {
        match self {
            MenuSelection::Stats => 0,
            MenuSelection::Quit  => 1,
        }
    }
}

pub struct ListSelection {
    pub selection: Option<usize>
}

impl ListSelection {
    pub fn increment(&mut self, max: usize) {
        if let Some(cur_selection) = self.selection {
            self.selection = Some({
                if cur_selection + 1 >= max {
                    0
                }

                else {
                    cur_selection + 1
                }
            });
        }
    }

    pub fn decrement(&mut self, max: usize) {
        if let Some(mut cur_selection) = self.selection {
            self.selection = Some({
                cur_selection = cur_selection - 1;
                if cur_selection > max - 1 {
                        max - 1
                }

                else {
                    cur_selection
                }
            });
        }
    }

    pub fn activate(&mut self) {
        self.selection = Some(0);
    }

    pub fn deactivate(&mut self) {
        self.selection = None;
    }

    pub fn is_active(&self) -> bool {
        self.selection.is_some()
    }
}

impl Default for ListSelection {
    fn default() -> Self {
        Self{selection: None}
    }
}

pub struct ConsoleUI {
    data:              ConsoleUIData,
    rx:                EventReceiver,
    terminal:          Terminal,
    console_size:      Rect,
    menu_selection:    MenuSelection,
    section_selection: ListSelection,
    first_section_id:  usize,
    second_section_id: usize,
}

impl ConsoleUI {
    pub fn new(rx: EventReceiver, terminal: Terminal) -> ConsoleUI {
        ConsoleUI{data: ConsoleUIData::default(), rx, terminal, console_size: Rect::default(), menu_selection: MenuSelection::Stats, section_selection: ListSelection::default(), first_section_id: 0, second_section_id: 0}
    }

    pub fn update_data(&mut self, memory_map: &MemoryMap) -> Result<(), Error> {
        self.data = ConsoleUIData::from(memory_map)?;

        if self.data.section_info.len() > 0 {
            if !self.section_selection.is_active() {
                self.section_selection.activate();
            }
        }

        else {
            self.section_selection.deactivate();
        }
        Ok(())
    }

    pub fn update(&mut self) -> Result<UIState, Error> {
        let input_result = {
            match self.rx.recv()? {
                Event::Input(event) => {
                    match event.code {
                        KeyCode::Char('q') => Ok(UIState::Terminated),
                        KeyCode::Char('s') => {
                            self.menu_selection = MenuSelection::Stats;
                            Ok(UIState::Render)
                        },
                        KeyCode::Left => {
                            self.menu_selection = self.menu_selection.prev();
                            Ok(UIState::Render)
                        },
                        KeyCode::Right => {
                            self.menu_selection = self.menu_selection.next();
                            Ok(UIState::Render)
                        },
                        KeyCode::Down => {
                            if matches!(self.menu_selection, MenuSelection::Stats) {
                                self.section_selection.increment(self.data.section_info.len());
                                Ok(UIState::Render)
                            }

                            else {
                                Ok(UIState::Alive)
                            }
                        },
                        KeyCode::Up => {
                            if matches!(self.menu_selection, MenuSelection::Stats) {
                                self.section_selection.decrement(self.data.section_info.len());
                                Ok(UIState::Render)
                            }

                            else {
                                Ok(UIState::Alive)
                            }
                        },
                        KeyCode::Enter => {
                            if matches!(self.menu_selection, MenuSelection::Quit) {
                                Ok(UIState::Terminated)
                            }
                            else {
                                Ok(UIState::Alive)
                            }
                        },
                        _ => Ok(UIState::Alive)
                    }
                },
                Event::Tick => {
                    let terminal_size = self.terminal.size().expect("Getting size of terminal");
                    if terminal_size != self.console_size {
                        self.console_size = terminal_size;
                        Ok(UIState::Render)
                    } 

                    else {
                        Ok(UIState::Alive)
                    }
                },
                Event::ForceRender => Ok(UIState::Render)
            }
        };

        self.first_section_id = {
            if let Some(selection) = self.section_selection.selection {
                selection
            }

            else {
                0
            }
        };

        input_result
    }

    pub fn render(&mut self) -> Result<(), Error> {
        self.terminal.draw(|frame| {
            // Create layout elements
            let main_layout = Layout::default().direction(Direction::Vertical).constraints([
                Constraint::Length(3),
                Constraint::Min(3)
            ]).split(frame.size()).to_vec();

            // Elements
            let menu_bar = {
                const MENU_TITLES: &'static [&'static str] = &["Stats", "Quit"];

                let menu = MENU_TITLES.iter().map(|t| {
                    let (first, rest) = t.split_at(1);
                    Line::from(vec![
                        Span::styled(first, Style::default().fg(Color::Yellow).add_modifier(Modifier::UNDERLINED),),
                        Span::styled(rest,  Style::default().fg(Color::White)),
                    ])
                }).collect();
                
                Tabs::new(menu).select(self.menu_selection.get_id()).block(
                    Block::default()
                    .title("psxreadmap - Menu").borders(Borders::ALL))
                    .style(Style::default().fg(Color::White))
                    .highlight_style(Style::default().bg(Color::LightYellow))
                    .divider(Span::raw("|")
                )
            };

            // Render screen
            frame.render_widget(menu_bar, main_layout[0]);
            match self.menu_selection {
                MenuSelection::Stats => {
                    // ToDo:                                   v This is the same and we always have elements now
                    Self::render_stats(frame, main_layout[1], (&self.section_selection, self.first_section_id), &mut self.data);
                },
                MenuSelection::Quit  => {
                    frame.render_widget(Self::create_exit_message(), main_layout[1]);
                }
            }
        })?;
        Ok(())
    }

    pub fn close(mut self) -> Result<(), Error> {
        crossterm::terminal::disable_raw_mode()?;
        self.terminal.show_cursor()?;
        self.terminal.clear()?;

        Ok(())
    }

    fn render_stats(frame: &mut ConsoleFrame, layout: Rect, content_selection: (&ListSelection, usize), data: &mut ConsoleUIData) {
        let layout = Layout::default().direction(Direction::Vertical).constraints([
            Constraint::Length(3),
            Constraint::Max(3),
            Constraint::Max(3),
        ]).split(layout).to_vec();
        let content_list = List::new({
            let mut vec = Vec::new();

            for section in &data.section_info {
                vec.push(ListItem::new(section.name.clone()))
            }
            vec
        }).style(Style::default().fg(Color::White))
        .block(Block::default()
            .borders(Borders::ALL)
            .style(Style::default().fg(Color::White))
            .border_type(BorderType::Plain)
        )
        .highlight_style(Style::default().bg(Color::Cyan));

        let first_mem_gauge  = Self::create_mem_gauge_from(&data.section_info[content_selection.1], Color::Cyan);
        let second_mem_gauge = Self::create_mem_gauge_from(&data.section_info[0], Color::Red);
        let mut list_state = ListState::default();
        
        list_state.select(content_selection.0.selection);

        frame.render_stateful_widget(content_list, layout[0], &mut list_state);
        frame.render_widget(first_mem_gauge, layout[1]);
        frame.render_widget(second_mem_gauge, layout[2]);
    }

    fn create_mem_gauge_from<'a>(section_info: &SectionInfo, color: Color) -> Gauge<'a> {
        let adr_ratio = section_info.get_size_ratio();
        Gauge::default().block(Block::default().borders(Borders::ALL)
            .title(format!("{} @0x{:x} - 0x{:x} [{}b/{}b]", 
                section_info.name, section_info.start_adr, section_info.get_end_adr(), section_info.size, section_info.max_size)))
            .gauge_style(Style::default().fg(color).bg(Color::Black).add_modifier(Modifier::BOLD))
            .ratio(adr_ratio).label(format!("{}%", adr_ratio*100.0f64))
    }

    fn create_exit_message<'a>() -> Paragraph<'a> {
        Paragraph::new("Press \"ENTER\" to exit")
            .style(Style::default().fg(Color::White))
            .alignment(Alignment::Center)
            .block(Block::default()
                .borders(Borders::ALL)
                .style(Style::default().fg(Color::White))
                .border_type(BorderType::Plain).title("QUIT")
            )
    }
}

#[derive(Default)]
struct ConsoleUIData {
    section_info: Vec<SectionInfo>
}

impl ConsoleUIData {
    pub const RAM_BASE_ADDRESS:u64    = 0x80000000;
    pub const RAM_SIZE:usize          = 2*1024*1024;

    pub fn from(memory_map: &MemoryMap) -> Result<Self, Error> {
        fn get_last_section_end_adr(memory_map: &MemoryMap) -> u64 {
            if let Some(section) = memory_map.sections.last() {
                return section.adr + section.size as u64;
            }
            0u64
        }

        let mut new_self     = ConsoleUIData::default();
        let     highest_adr  = get_last_section_end_adr(memory_map);
        let     program_size = (highest_adr - Self::RAM_BASE_ADDRESS) as usize;

        new_self.section_info.push(SectionInfo{
            name:      String::from("<overall usage>"),
            start_adr: Self::RAM_BASE_ADDRESS,
            size:      program_size,
            max_size:  Self::RAM_SIZE
        });

        for section in &memory_map.sections {
            new_self.section_info.push(SectionInfo{
                name:      section.name.clone(),
                start_adr: section.adr,
                size:      section.size,
                max_size:  program_size
            });
        }

        Ok(new_self)
    }
}

#[derive(Default)]
struct SectionInfo {
    pub name:      String,
    pub start_adr: u64,
    pub size:      usize,
    pub max_size:  usize,
}

impl SectionInfo {
    fn get_end_adr(&self) -> u64 {
        self.start_adr + self.size as u64
    }

    fn get_size_ratio(&self) -> f64 {
        self.size as f64 / self.max_size as f64
    }
}

pub fn load_memory_map(use_wsl: bool, input: PathBuf) -> Result<MemoryMap, Error> {
    readmap_helper::generate_memory_map(use_wsl, input)
}